/* xgetopt.c - simple getopt and getopt_long alternative
 *
 * Copyright (C) 2013, 2020 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * gcc -D TEST xgetopt.c -o xgetopt-test
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#ifndef XUTIL_H
#include "xutil.h"
#endif

#ifndef XGETOPT_H
#include "xgetopt.h"
#endif


int
xgetopt (int argc, char *argv[], int *idx, XGetOption *option, int n_options)
{
	int opt, len;
	xbool_t not_last;
	const char *str;


	str = argv[*idx];

	if (*str != '-') {
		return XGETOPT_NONE;
	}

	*idx = *idx + 1;

	str++;
	if (*str == '-') {
		str++;
	}
	len = strlen (str);

	not_last = *idx < argc;

	if (len > 0) {
		for (opt = 0; opt < n_options; opt++) {
			if (len >= option[opt].min_len &&
				strncmp (str, option[opt].opt_str, len) == 0) {
				if (!option[opt].requires_arg || not_last) {
					return opt;
				}
				else {
					return XGETOPT_MISSING_ARG;
				}
			}
		}
	}
	return XGETOPT_UNKNOWN;
}


#ifdef TEST

static XGetOption option[] = {
	{1, TRUE,  "foo"},
	{1, TRUE,  "bar"},
	{1, FALSE, "help"}
};

enum {
	OPT_FOO,
	OPT_BAR,
	OPT_HELP,
	N_OPTS
};


static void
show_usage (const char *progname)
{
	printf ("\nUsage: %s [OPTIONS] ...\n\n", progname);
	printf ("Options:\n\n");
	printf ("  -b, --bar VALUE   Give a value for bar\n");
	printf ("  -f, --foo VALUE   Give a value for foo\n");
	printf ("  -h, --help        Show this usage information\n");
	printf ("\n");
}


int
main (int argc, char *argv[])
{
	int opt, idx;
	const char *foo, *bar;

	foo = bar = NULL;

	opt = 1;
	idx = 1;

	while (idx < argc && opt > XGETOPT_NONE) {

		opt = xgetopt (argc, argv, &idx, option, N_OPTS);

		switch (opt) {

		case OPT_HELP:
			show_usage (argv[0]);
			exit (0);
			break;

		case OPT_FOO:
			foo = argv[idx++];
			break;

		case OPT_BAR:
			bar = argv[idx++];
			break;

		case XGETOPT_MISSING_ARG:
			fprintf (stderr, "option '%s' requires an argument\n",
			         argv[idx - 1]);
			exit (1);

		case XGETOPT_UNKNOWN:
			fprintf (stderr, "unknown option '%s'\n", argv[idx - 1]);
			exit (1);

		default:
			/* assert - shouldn't get here */
			break;

		}

	}

	printf ("foo = %s, bar = %s, idx = %d, argc = %d\n",
	        foo, bar, idx, argc);

	while (idx < argc) {
		printf ("%s\n", argv[idx]);
		idx++;
	}

	return 0;
}

#endif

