/* xtree.c - binary search tree (not balanced)
 *
 * Copyright (C) 2009, 2020 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * gcc -D TEST xmem.o xstr.o xtree.c -o xtree-test
 *
 */


#include <stdio.h>

#if TEST
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#endif

#include <assert.h>

#ifndef XUTIL_H
#include "xutil.h"
#endif

#ifndef XMEM_H
#include "xmem.h"
#endif

#ifndef XTREE_H
#include "xtree.h"
#endif

#ifdef TEST
#include "xstr.h"
#endif


typedef struct _xnode xnode_t;

struct _xnode {
	void    *key;
	void    *data;
	xnode_t *left;
	xnode_t *right;
};

struct _xtree_t {
	xnode_t *root;
	XCmp_f  cmp;        /* Compare keys */
	XFree_f free_key;   /* Free keys (otherwise NULL) */
	XFree_f free_data;  /* Free data (otherwise NULL) */
};


static xnode_t* xnode_new (void *key, void *data);
static void     xnode_free (xtree_t *t, xnode_t *tn);
static xnode_t* xnode_insert (xtree_t *t, xnode_t *tn, void *key, void *data,
                              xbool_t replace);
static void*    xnode_lookup (xtree_t *t, xnode_t *tn, void *key);
static xnode_t* xnode_remove (xtree_t *t, xnode_t *tn, void *key);
static xnode_t* xnode_leftmost (xnode_t *tn);
static int      xnode_n_items (xnode_t *tn);
static int      xnode_height (xnode_t *tn);
static int      xnode_shortest_path (xnode_t *tn);
static xbool_t  xnode_foreach (xnode_t *tn, XTree_f func, void *user_data);


static xnode_t*
xnode_new (void *key, void *data)
{
	xnode_t *tn;

	tn = xnew (xnode_t, 1);
	tn->key = key;
	tn->data = data;
	tn->left = NULL;
	tn->right = NULL;

	return tn;
}


/* Free all nodes allocated for tree tn.  */

static void
xnode_free (xtree_t *t, xnode_t *tn)
{
	if (tn != NULL) {

		xnode_free (t, tn->left);
		xnode_free (t, tn->right);

		if (t->free_key != NULL) {
			t->free_key (tn->key);
		}
		if (t->free_data != NULL) {
			t->free_data (tn->data);
		}

		xfree (tn);
	}
}


/* Add a key/data pair to a tree.  */

static xnode_t*
xnode_insert (xtree_t *t, xnode_t *tn, void *key, void *data, xbool_t replace)
{
	int i;

	if (tn == NULL) {
		return xnode_new (key, data);
	}

	i = t->cmp (key, tn->key);

	if (i == 0) {

		/* Item exists, so replace its data. */

		if (t->free_data != NULL) {
			t->free_data (tn->data);
		}
		tn->data = data;

		if (replace) {

			/* Replace its key as well. */
			if (t->free_key != NULL) {
				t->free_key (tn->key);
			}
			tn->key = key;

		}
		else {

			/* Free the caller's key. */
			if (t->free_key != NULL) {
				t->free_key (key);
			}

		}

	}
	else if (i < 0) {
		tn->left = xnode_insert (t, tn->left, key, data, replace);
	}
	else {
		tn->right = xnode_insert (t, tn->right, key, data, replace);
	}
	return tn;
}


/* Return the data associated with a key.  */

static void*
xnode_lookup (xtree_t *t, xnode_t *tn, void *key)
{
	int i;

	if (tn == NULL) {
		return NULL;
	}

	i = t->cmp (key, tn->key);

	if (i == 0) {
		return tn->data;
	}
	else if (i < 0) {
		return xnode_lookup (t, tn->left, key);
	}
	else {
		return xnode_lookup (t, tn->right, key);
	}
}


/* Return the leftmost node (lowest value).  */

static xnode_t*
xnode_leftmost (xnode_t *tn)
{
	assert (tn != NULL);

	while (tn->left != NULL) {
		tn = tn->left;
	}
	return tn;
}


/* Remove a node from a tree, and return it.  */

static xnode_t*
xnode_remove (xtree_t *t, xnode_t *tn, void *key)
{
	if (tn == NULL) {
		return NULL;
	}

	if (t->cmp (key, tn->key) == 0) {

		if (tn->left != NULL && tn->right != NULL) {

			xnode_t *leftmost;

			if (t->free_key != NULL) {
				t->free_key (tn->key);
			}
			if (t->free_data != NULL) {
				t->free_data (tn->data);
			}

			leftmost = xnode_leftmost (tn->right);

			tn->key = leftmost->key;
			tn->data = leftmost->data;

			key = tn->key;

		}
		else {

			xnode_t *tmp = tn;

			if (tn->left != NULL) {
				tn = tn->left;
			}
			else { 
				tn = tn->right;
			}

			xfree (tmp);
			
			return tn;
		}

	}

	if (t->cmp (key, tn->key) < 0) {
		tn->left = xnode_remove (t, tn->left, key);
	}
	else {
		tn->right = xnode_remove (t, tn->right, key);
	}

	return tn;
}


/* Return the number of nodes in tree tn.  */

static int
xnode_n_items (xnode_t *tn)
{
	if (tn == NULL) {
		return 0;
	}
	return 1 + xnode_n_items (tn->left) + xnode_n_items (tn->right);
}


/* Return the height of tree tn.  */

static int
xnode_height (xnode_t *tn)
{
	if (tn == NULL) {
		return 0;
	}
	return 1 + MAX (xnode_height (tn->left), xnode_height (tn->right));
}


/* Return the length of the shortest path in tree tn.  */

static int
xnode_shortest_path (xnode_t *tn)
{
	if (tn == NULL) {
		return 0;
	}
	return 1 + MIN (xnode_shortest_path (tn->left),
	                xnode_shortest_path (tn->right));
}


/* Perform a function on each element (inorder) of tree tn.
   Stop if the function returns TRUE.  */

static xbool_t
xnode_foreach (xnode_t *tn, XTree_f func, void *user)
{
	if (tn != NULL) {
		if (xnode_foreach (tn->left, func, user)) {
			return TRUE;
		}
		if (func (tn->key, tn->data, user) == TRUE) {
			return TRUE;
		}
		if (xnode_foreach (tn->right, func, user)) {
			return TRUE;
		}
	}
	return FALSE;
}


xtree_t*
xtree_new (XCmp_f cmp, XFree_f free_key, XFree_f free_data)
{
	xtree_t *t;

	assert (cmp != NULL);

	t = xnew (xtree_t, 1);
	t->root = NULL;
	t->cmp = cmp;
	t->free_key = free_key;
	t->free_data = free_data;

	return t;
}


void
xtree_free (xtree_t *t)
{
	assert (t != NULL);
	xnode_free (t, t->root);
	xfree (t);
}


void
xtree_insert (xtree_t *t, void *key, void *data)
{
	assert (t != NULL);
	t->root = xnode_insert (t, t->root, key, data, FALSE);
}


void
xtree_replace (xtree_t *t, void *key, void *data)
{
	assert (t != NULL);
	t->root = xnode_insert (t, t->root, key, data, TRUE);
}


void*
xtree_lookup (xtree_t *t, void *key)
{
	assert (t != NULL);
	return xnode_lookup (t, t->root, key);
}


void
xtree_remove (xtree_t *t, void *key)
{
	assert (t != NULL);
	t->root = xnode_remove (t, t->root, key);
}


int
xtree_n_items (xtree_t *t)
{
	assert (t != NULL);
	return xnode_n_items (t->root);
}


int
xtree_height (xtree_t *t)
{
	assert (t != NULL);
	return xnode_height (t->root);
}


int
xtree_shortest_path (xtree_t *t)
{
	assert (t != NULL);
	return xnode_shortest_path (t->root);
}


void
xtree_foreach (xtree_t *t, XTree_f func, void *user)
{
	assert (t != NULL);
	xnode_foreach (t->root, func, user);
}


#ifdef TEST

static void
preorder (xnode_t *tn)
{
	if (tn != NULL) {
		printf ("  '%ld' -> '%s'\n", (long)tn->key, (char*)tn->data);
		preorder (tn->left);
		preorder (tn->right);
	}
}


static void
inorder (xnode_t *tn)
{
	if (tn != NULL) {
		inorder (tn->left);
		printf ("  '%ld' -> '%s'\n", (long)tn->key, (char*)tn->data);
		inorder (tn->right);
	}
}


static void
postorder (xnode_t *tn)
{
	if (tn != NULL) {
		postorder (tn->left);
		postorder (tn->right);
		printf ("  '%ld' -> '%s'\n", (long)tn->key, (char*)tn->data);
	}
}


static void
xnode_print (xnode_t *tn, int level)
{
	int i;

	if (tn == NULL) {
		for (i = 0; i < level; i++) {
			putchar ('\t');
		}
		puts ("~");
	}
	else {
		xnode_print (tn->left, level + 1);
		for (i = 0; i < level; i++) {
			putchar ('\t');
		}
		printf ("%ld\n", (long)tn->key);
		xnode_print (tn->right, level + 1);
	}
}


static int
n_leaves (xnode_t *tn)
{
	if (tn == NULL) {
		return 0;
	}
	if (tn->left == NULL && tn->right == NULL) {
		return 1;
	}
	return n_leaves (tn->left) + n_leaves (tn->right);
}


static int
my_cmp (void *data1, void *data2)
{
	long a, b;

	a = (long)data1;
	b = (long)data2;

	if (a == b) {
		return 0;
	}
	else if (a < b) {
		return -1;
	}
	return 1;
}


int
main (int argc, char *argv[])
{
	xtree_t *t;
	void *s;


	t = xtree_new ((XCmp_f)my_cmp, NULL, NULL);

	xtree_insert (t, (void*)25, "Apple");
	xtree_insert (t, (void*)13, "Pear");
	xtree_insert (t, (void*)33, "Banana");
	xtree_insert (t, (void*)5, "Apricot");
	xtree_insert (t, (void*)18, "Peach");
	xtree_insert (t, (void*)27, "Orange");
	xtree_insert (t, (void*)15, "Cherry");

	printf ("n items:       %d\n", xtree_n_items (t));
	printf ("height:        %d\n", xtree_height (t));
	printf ("shortest path: %d\n", xtree_shortest_path (t));
	printf ("n leaves:      %d\n", n_leaves (t->root));

	printf ("inorder:\n");
	inorder (t->root);

#if 0
	s = (void*)25;
	printf ("remove '%ld'\n", (long)s);
	xtree_remove (t, s);
#endif

	puts ("preorder:");
	preorder (t->root);

	puts ("inorder:");
	inorder (t->root);

	puts ("postorder:");
	postorder (t->root);

	s = (void*)33;
	printf ("lookup %ld: %s\n", (long)s, (char*)xtree_lookup (t, s));

	xnode_print (t->root, 0);
	putchar ('\n');

	xtree_free (t);

	return EXIT_SUCCESS;
}

#endif /* TEST */

